#include <stdlib.h>
#include <stdio.h>
#include <pthread.h>
#include <assert.h>
#include <getopt.h>
#include <limits.h>
#include <signal.h>
#include <sys/time.h>
#include <time.h>

#include "utils.h"


#define TAS    0
#define TTAS   1
#define TICKET 2

#define USE_CAS      0
#define USE_FAI      1
#define USE_SWAP     2
#define USE_FAI_CAS  3
#define USE_SWAP_CAS 4

// use https://gcc.gnu.org/onlinedocs/gcc-4.4.5/gcc/Atomic-Builtins.html
// for all but SWAP
// for swap use the following:

//Swap uint64_t
inline uint64_t swap_uint64(volatile uint64_t* target,  uint64_t x);
//Swap uint32_t
inline uint32_t swap_uint32(volatile uint32_t* target,  uint32_t x);
//Swap uint16_t
inline uint16_t swap_uint16(volatile uint16_t* target,  uint16_t x);
//Swap uint8_t
inline uint8_t swap_uint8(volatile uint8_t* target,  uint8_t x);

#define COMPILER_BARRIER()   asm volatile("" ::: "memory");

#define CAS(m, old, new)          __sync_val_compare_and_swap(m, old, new)

#if ATOMIC_TYPE == USE_FAI_CAS
#  define FAI(m) 
#elif ATOMIC_TYPE == USE_SWAP_CAS
#  define SWAP(m) 
#else
#  define FAI(m) 		/* normal FAI */
#  define SWAP(m) 		/* noraml SWAP */
#endif

#if LOCK_TYPE == TAS /* TAS ***************************************************/
typedef struct
{
  volatile uint64_t lock;
  // lock data TAS -- volatile?
} lock_t;
#elif LOCK_TYPE == TTAS	/* TTAS ***************************************************/
typedef struct
{
  // lock data TTAS -- volatile?
} lock_t;
#elif LOCK_TYPE == TICKET /* TICKET ***************************************************/
typedef struct
{
  // lock data TICKET -- volatile?
} lock_t;
#endif

#if LOCK_TYPE == TAS /* TAS ***************************************************/
void
lock_init(lock_t* l)
{
  l->lock = 0;
}

void
lock(lock_t* l)
{
#if ATOMIC_TYPE == USE_CAS
  while (CAS(&l->lock, 0, 1) == 1)
    {
      asm volatile ("pause");
    }
#elif ATOMIC_TYPE == USE_FAI

#elif ATOMIC_TYPE == USE_SWAP 

#elif ATOMIC_TYPE == USE_FAI_CAS

#elif ATOMIC_TYPE == USE_SWAP_CAS

#endif
}

void
unlock(lock_t* l)
{
  COMPILER_BARRIER();  
  l->lock = 0;
  COMPILER_BARRIER();  
}


#elif LOCK_TYPE == TTAS /* TTAS ***************************************************/
void
lock_init(lock_t* l)
{
  
}

void
lock(lock_t* l)
{
#if ATOMIC_TYPE == USE_CAS

#elif ATOMIC_TYPE == USE_FAI

#elif ATOMIC_TYPE == USE_SWAP 

#elif ATOMIC_TYPE == USE_FAI_CAS

#elif ATOMIC_TYPE == USE_SWAP_CAS

#endif
}

void
unlock(lock_t* l)
{

}


#elif LOCK_TYPE == TICKET /* TICKET ***************************************************/
lock_init(lock_t* l)
{

}

void
lock(lock_t* l)
{

}

void
unlock(lock_t* l)
{

}
#endif

lock_t global_lock;

volatile size_t global_counter = 0;


#define DEFAULT_NUM_THREADS 2
#define DEFAULT_DURATION 1000

int duration;
int num_threads;

//static volatile int stop;

//used to signal the threads when to stop
ALIGNED(64) volatile int running[1];

//per-thread seeds for the custom random function
__thread unsigned long * seeds;

#define XSTR(s)                         STR(s)
#define STR(s)                          #s

//a simple barrier implementation
//used to make sure all threads start the experiment at the same time
typedef struct barrier
{
  pthread_cond_t complete;
  pthread_mutex_t mutex;
  int count;
  int crossing;
} barrier_t;

void barrier_init(barrier_t *b, int n)
{
    pthread_cond_init(&b->complete, NULL);
    pthread_mutex_init(&b->mutex, NULL);
    b->count = n;
    b->crossing = 0;
}

void barrier_cross(barrier_t *b)
{
    pthread_mutex_lock(&b->mutex);
    /* One more thread through */
    b->crossing++;
    /* If not all here, wait */
    if (b->crossing < b->count) {
        pthread_cond_wait(&b->complete, &b->mutex);
    } else {
        pthread_cond_broadcast(&b->complete);
        /* Reset for next time */
        b->crossing = 0;
    }
    pthread_mutex_unlock(&b->mutex);
}

//data structure through which we send parameters to and get results from the worker threads
typedef ALIGNED(64) struct thread_data {
    //pointer to the global barrier
    barrier_t *barrier;
    //counts the number of operations each thread performs
    unsigned long num_operations;
    //the number of elements each thread should add at the beginning of its execution
    uint64_t num_add;
    //number of inserts a thread performs
    unsigned long num_insert;
    //number of removes a thread performs
    unsigned long num_remove;
    //number of searches a thread performs
    unsigned long num_search;
    //the id of the thread (used for thread placement on cores)
    int id;
} thread_data_t;

void *test(void *data)
{
  //get the per-thread data
  thread_data_t *d = (thread_data_t *)data;
  //scale percentages of the various operations to the range 0..255
  //this saves us a floating point operation during the benchmark
  //e.g instead of random()%100 to determine the next operation we will do, we can simply do random()&256
  //this saves time on some platfroms
  /* uint32_t read_thresh = 256 * finds / 100; */
  uint32_t rand_max;
  //seed the custom random number generator
  seeds = seed_rand();
  volatile uint64_t index, sum = 0;
  /* Wait on barrier */
  barrier_cross(d->barrier);
  size_t n_ops = 0;
  //start the test
  while (*running)
    {
      lock(&global_lock);
      global_counter++;
      n_ops++;
      unlock(&global_lock);
    }

  d->num_operations = n_ops;

  return NULL;
}

void catcher(int sig)
{
    static int nb = 0;
    printf("CAUGHT SIGNAL %d\n", sig);
    if (++nb >= 3)
        exit(1);
}

int
main(int argc, char* const argv[])
{
  pthread_t *threads;
  pthread_attr_t attr;
  barrier_t barrier;
  struct timeval start, end;
  struct timespec timeout;

  thread_data_t *data;
  sigset_t block_set;

  //initially, set parameters to their default values
  num_threads = DEFAULT_NUM_THREADS;
  duration=DEFAULT_DURATION;

  //now read the parameters in case the user provided values for them 
  //we use getopt, the same skeleton may be used for other bechmarks,
  //though the particular parameters may be different
  struct option long_options[] = {
    // These options don't set a flag
    {"help",                      no_argument,       NULL, 'h'},
    {"duration",                  required_argument, NULL, 'd'},
    {"num-threads",               required_argument, NULL, 'n'},
    {NULL, 0, NULL, 0}
  };

  int i,c;

  //actually get the parameters form the command-line
  while(1) {
    i = 0;
    c = getopt_long(argc, argv, "hd:n:", long_options, &i);

    if(c == -1)
      break;

    if(c == 0 && long_options[i].flag == 0)
      c = long_options[i].val;

    switch(c) {
    case 0:
      /* Flag is automatically set */
      break;
    case 'h':
      printf("lock stress test\n"
	     "\n"
	     "Usage:\n"
	     "  stress_test [options...]\n"
	     "\n"
	     "Options:\n"
	     "  -h, --help\n"
	     "        Print this message\n"
	     "  -d, --duration <int>\n"
	     "        Test duration in milliseconds (0=infinite, default=" XSTR(DEFAULT_DURATION) ")\n"
	     "  -n, --num-threads <int>\n"
	     "        Number of threads (default=" XSTR(DEFAULT_NUM_THREADS) ")\n"
	     );
      exit(0);
    case 'd':
      duration = atoi(optarg);
      break;
    case 'n':
      num_threads = atoi(optarg);
      break;
    case '?':
      printf("Use -h or --help for help\n");
      exit(0);
    default:
      exit(1);
    }
  }

  //initialization of the lock
  lock_init(&global_lock);

  //initialize the data which will be passed to the threads
  if ((data = (thread_data_t *)malloc(num_threads * sizeof(thread_data_t))) == NULL) {
    perror("malloc");
    exit(1);
  }

  if ((threads = (pthread_t *)malloc(num_threads * sizeof(pthread_t))) == NULL) {
    perror("malloc");
    exit(1);
  }

  //flag signaling the threads until when to run
  printf("## Start the threads\n");
  *running = 1;

  //global barrier initialization (used to start the threads at the same time)
  barrier_init(&barrier, num_threads + 1);
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

  timeout.tv_sec = duration / 1000;
  timeout.tv_nsec = (duration % 1000) * 1000000;
    

  //set the data for each thread and create the threads
  for (i = 0; i < num_threads; i++) {
    data[i].id = i;
    data[i].num_operations = 0;
    data[i].num_insert=0;
    data[i].num_remove=0;
    data[i].num_search=0;
    data[i].barrier = &barrier;
    if (pthread_create(&threads[i], &attr, test, (void *)(&data[i])) != 0) {
      fprintf(stderr, "Error creating thread\n");
      exit(1);
    }
  }
  pthread_attr_destroy(&attr);

  /* Catch some signals */
  if (signal(SIGHUP, catcher) == SIG_ERR ||
      signal(SIGINT, catcher) == SIG_ERR ||
      signal(SIGTERM, catcher) == SIG_ERR) {
    perror("signal");
    exit(1);
  }

  /* Start threads */
  barrier_cross(&barrier);
  gettimeofday(&start, NULL);
  if (duration > 0) {
    //sleep for the duration of the experiment
    nanosleep(&timeout, NULL);
  } else {
    sigemptyset(&block_set);
    sigsuspend(&block_set);
  }

  //signal the threads to stop
  printf("## Stop the threads\n");
  *running = 0;
  gettimeofday(&end, NULL);

  /* Wait for thread completion */
  for (i = 0; i < num_threads; i++) {
    if (pthread_join(threads[i], NULL) != 0) {
      fprintf(stderr, "Error waiting for thread completion\n");
      exit(1);
    }
  }
  //compute the exact duration of the experiment
  duration = (end.tv_sec * 1000 + end.tv_usec / 1000) - (start.tv_sec * 1000 + start.tv_usec / 1000);
    
  unsigned long operations = 0;
  long reported_total = 0; 
  //report some experiment statistics
  for (i = 0; i < num_threads; i++) {
    operations += data[i].num_operations;
    reported_total = reported_total + data[i].num_add + data[i].num_insert - data[i].num_remove;
  }

  /* printf("Duration      : %d (ms)\n", duration); */
  double thr = operations * 1000.0 / duration;
  double thr_Mops = thr / (1024 * 1024);
  printf("Throughput: %.2f Mop/s\n", thr_Mops);
  printf("# global counter: %-10zu vs. \n# of increments : %-10zu\n", global_counter, operations);
  if (global_counter == operations)
    {
      printf(" :-) :-) correct, atomic counter!\n");
      printf(" ------------------------------------------------------------------------- TEST PASSED\n");
    }
  else
    {
      printf(" :-( :-( incorrect, non-atomic counter!\n");
      printf(" (your lock / unlock / init implemntation is broken)\n");
    }
    
  free(threads);
  free(data);
  return 0;

}

//Swap uint64_t
inline uint64_t swap_uint64(volatile uint64_t* target,  uint64_t x) {
  __asm__ __volatile__("xchgq %0,%1"
        :"=r" ((uint64_t) x)
        :"m" (*(volatile uint64_t *)target), "0" ((uint64_t) x)
        :"memory");

  return x;
}

//Swap uint32_t
inline uint32_t swap_uint32(volatile uint32_t* target,  uint32_t x) {
  __asm__ __volatile__("xchgl %0,%1"
        :"=r" ((uint32_t) x)
        :"m" (*(volatile uint32_t *)target), "0" ((uint32_t) x)
        :"memory");

  return x;
}

//Swap uint16_t
inline uint16_t swap_uint16(volatile uint16_t* target,  uint16_t x) {
  __asm__ __volatile__("xchgw %0,%1"
        :"=r" ((uint16_t) x)
        :"m" (*(volatile uint16_t *)target), "0" ((uint16_t) x)
        :"memory");

  return x;
}

//Swap uint8_t
inline uint8_t swap_uint8(volatile uint8_t* target,  uint8_t x) {
  __asm__ __volatile__("xchgb %0,%1"
        :"=r" ((uint8_t) x)
        :"m" (*(volatile uint8_t *)target), "0" ((uint8_t) x)
        :"memory");

  return x;
}
